#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int convert_index(int number)
{
	// add together all the numbers below number which are divisible by the obvious primes
	int set = number/2 + number/3 + number/5 + number/7;

	// subtract the overlap
	set -= number/(2*3) + number/(2*5) + number/(2*7) + number/(3*5) + +number/(3*7) + number/(5*7);

	// add the overlap's overlap
	set += number/(2*3*5) + number/(2*3*7) + number/(2*5*7) + number/(3*5*7);

	// finally subtract the overlap's overlap's overlap
	set -= number/(2*3*5*7);

	return number - set - 2;
}

int is_prime_file(int number, FILE *fp)
{
	// remove the obvious first
	if(number == 2 || number == 3 || number == 5 || number == 7)
		return 1;
	if(number < 2 || number%2 == 0 || number%3 == 0 || number%5 == 0 || number%7 == 0)
		return 0;

	number = convert_index(number);

	int real_index = number/8;
	int bit_index = number%8;

	fseek(fp, real_index, 0);
	unsigned char primes = fgetc(fp);

	return primes & (1<<bit_index);
}

int is_prime_buffer(int number, unsigned char *buffer)
{
	// remove the obvious first
	if(number == 2 || number == 3 || number == 5 || number == 7)
		return 1;
	if(number < 2 || number%2 == 0 || number%3 == 0 || number%5 == 0 || number%7 == 0)
		return 0;

	number = convert_index(number);

	int real_index = number/8;
	int bit_index = number%8;

	return buffer[real_index] & (1<<bit_index);
}

void remove_multiples(unsigned char *buffer, int size, int number)
{
	int i;
	int index;
	int real_index;
	int bit_index;

	i=2*number;
	while((index = convert_index(i)) < size)
	{
		if(!(i%2 == 0 || i%3 == 0 || i%5 == 0 || i%7 == 0))
		{
			real_index = index/8;
			bit_index = index%8;
			buffer[real_index] &= ~(1<<bit_index);
		}
		i += number;
	}
}

void generate_primes(unsigned char *buffer, int size)
{
	int i;
	int max = size*8;

	// add together all the numbers below number which are divisible by the obvious primes
	int set = max/2 + max/3 + max/5 + max/7;

	// subtract the overlap
	set -= max/(2*3) + max/(2*5) + max/(2*7) + max/(3*5) + +max/(3*7) + max/(5*7);

	// add the overlap's overlap
	set += max/(2*3*5) + max/(2*3*7) + max/(2*5*7) + max/(3*5*7);

	// finally subtract the overlap's overlap's overlap
	set -= max/(2*3*5*7);

	max += set + 2;

	// set all the bits to 1
	memset(buffer, 0xFF, size);

	for(i=11; convert_index(i)<(int)sqrt(max); ++i)
	{
		if(is_prime_buffer(i, buffer))
		{
			remove_multiples(buffer, size*8, i);
		}
	}
}

int main(int argc, char *argv[])
{
	if(argc == 3)
	{
		int size = atoi(argv[2]);
		unsigned char *buffer = malloc(size);
		generate_primes(buffer, size);
		FILE *fp = fopen(argv[1], "wb+");
		fwrite(buffer, 1, size, fp);
		fclose(fp);
		free(buffer);
	}
	else if(argc == 2)
	{
		FILE *fp = fopen(argv[1], "rb");
		int number = 0;
		printf("Enter number to check if it's prime, enter -1 to exit\n");
		scanf("%d", &number);
		while(number != -1)
		{
			if(is_prime_file(number, fp))
			{
				printf("%d is prime\n", number);
			}
			else
			{
				printf("%d is not prime\n", number);
			}
			printf("Enter number to check if it's prime, enter -1 to exit\n");
			scanf("%d", &number);
		}

		fclose(fp);
	}
	else
	{
		printf("Usage:\n\n");
		printf("cprime <file> <size>\n");
		printf("If <size> is specified, a file will be generated containing prime numbers\n");
	}
	return 0;
}

